@@ -9,15 +9,16 @@ from zipfile import ZipFile
 from datetime import datetime
 from io import BytesIO
 import ssl
+from typing import List, Optional, Tuple, Union
 from urllib.parse import quote
 from urllib.request import urlopen, Request
 from urllib.error import HTTPError, URLError
 from termcolor import colored
 import colorama  # Required for Windows
-import argcomplete
+import shtab
 
-__version__ = "2.0.0"
-__client_specification__ = "1.4"
+__version__ = "3.1.0"
+__client_specification__ = "1.5"
 
 REQUEST_HEADERS = {'User-Agent': 'tldr-python-client'}
 PAGES_SOURCE_LOCATION = os.environ.get(
@@ -29,8 +30,9 @@ DOWNLOAD_CACHE_LOCATION = os.environ.get(
     'https://tldr-pages.github.io/assets/tldr.zip'
 )
 
+USE_NETWORK = int(os.environ.get('TLDR_NETWORK_ENABLED', '1')) > 0
 USE_CACHE = int(os.environ.get('TLDR_CACHE_ENABLED', '1')) > 0
-MAX_CACHE_AGE = int(os.environ.get('TLDR_CACHE_MAX_AGE', 24))
+MAX_CACHE_AGE = int(os.environ.get('TLDR_CACHE_MAX_AGE', 24*7))
 
 URLOPEN_CONTEXT = None
 if int(os.environ.get('TLDR_ALLOW_INSECURE', '0')) == 1:
@@ -50,7 +52,7 @@ class CacheNotExist(Exception):
     pass
 
 
-def get_language_code(language):
+def get_language_code(language: str) -> str:
     language = language.split('.')[0]
     if language in ['pt_PT', 'pt_BR', 'zh_TW']:
         return language
@@ -59,7 +61,7 @@ def get_language_code(language):
     return language.split('_')[0]
 
 
-def get_default_language():
+def get_default_language() -> str:
     default_lang = get_language_code(
         os.environ.get(
             'LANG',
@@ -73,22 +75,22 @@ def get_default_language():
     return default_lang
 
 
-def get_cache_dir():
-    if not os.environ.get('XDG_CACHE_HOME', False):
-        if not os.environ.get('HOME', False):
-            return os.path.join(os.path.expanduser("~"), ".cache", "tldr")
+def get_cache_dir() -> str:
+    if os.environ.get('XDG_CACHE_HOME', False):
+        return os.path.join(os.environ.get('XDG_CACHE_HOME'), 'tldr')
+    if os.environ.get('HOME', False):
         return os.path.join(os.environ.get('HOME'), '.cache', 'tldr')
-    return os.path.join(os.environ.get('XDG_CACHE_HOME'), 'tldr')
+    return os.path.join(os.path.expanduser("~"), ".cache", "tldr")
 
 
-def get_cache_file_path(command, platform, language):
+def get_cache_file_path(command: str, platform: str, language: str) -> str:
     pages_dir = "pages"
     if language and language != 'en':
         pages_dir += "." + language
     return os.path.join(get_cache_dir(), pages_dir, platform, command) + ".md"
 
 
-def load_page_from_cache(command, platform, language):
+def load_page_from_cache(command: str, platform: str, language: str) -> Optional[str]:
     try:
         with open(get_cache_file_path(
             command,
@@ -101,7 +103,12 @@ def load_page_from_cache(command, platform, language):
         pass
 
 
-def store_page_to_cache(page, command, platform, language):
+def store_page_to_cache(
+    page: str,
+    command: str,
+    platform: str,
+    language: str
+) -> Optional[str]:
     try:
         cache_file_path = get_cache_file_path(command, platform, language)
         os.makedirs(os.path.dirname(cache_file_path), exist_ok=True)
@@ -111,7 +118,7 @@ def store_page_to_cache(page, command, platform, language):
         pass
 
 
-def have_recent_cache(command, platform, language):
+def have_recent_cache(command: str, platform: str, language: str) -> bool:
     try:
         cache_file_path = get_cache_file_path(command, platform, language)
         last_modified = datetime.fromtimestamp(os.path.getmtime(cache_file_path))
@@ -121,7 +128,7 @@ def have_recent_cache(command, platform, language):
         return False
 
 
-def get_page_url(command, platform, remote, language):
+def get_page_url(command: str, platform: str, remote: str, language: str) -> str:
     if remote is None:
         remote = PAGES_SOURCE_LOCATION
 
@@ -133,7 +140,13 @@ def get_page_url(command, platform, remote, language):
     return remote + language + "/" + platform + "/" + quote(command) + ".md"
 
 
-def get_page_for_platform(command, platform, remote, language, only_use_cache=False):
+def get_page_for_platform(
+    command: str,
+    platform: str,
+    remote: str,
+    language: str,
+    only_use_cache: bool = False
+) -> str:
     data_downloaded = False
     if USE_CACHE and have_recent_cache(command, platform, language):
         data = load_page_from_cache(command, platform, language)
@@ -151,15 +164,22 @@ def get_page_for_platform(command, platform, remote, language, only_use_cache=Fa
             ).read()
             data_downloaded = True
         except Exception:
+            if not USE_CACHE:
+                raise
             data = load_page_from_cache(command, platform, language)
             if data is None:
                 raise
-    if data_downloaded:
+    if data_downloaded and USE_CACHE:
         store_page_to_cache(data, command, platform, language)
     return data.splitlines()
 
 
-def update_page_for_platform(command, platform, remote, language):
+def update_page_for_platform(
+    command: str,
+    platform: str,
+    remote: str,
+    language: str
+) -> None:
     page_url = get_page_url(platform, command, remote, language)
     data = urlopen(
         Request(page_url, headers=REQUEST_HEADERS),
@@ -168,14 +188,14 @@ def update_page_for_platform(command, platform, remote, language):
     store_page_to_cache(data, command, platform, language)
 
 
-def get_platform():
+def get_platform() -> str:
     for key in OS_DIRECTORIES:
         if sys.platform.startswith(key):
             return OS_DIRECTORIES[key]
     return 'linux'
 
 
-def get_platform_list():
+def get_platform_list() -> List[str]:
     platforms = ['common'] + list(OS_DIRECTORIES.values())
     current_platform = get_platform()
     platforms.remove(current_platform)
@@ -183,7 +203,7 @@ def get_platform_list():
     return platforms
 
 
-def get_language_list():
+def get_language_list() -> List[str]:
     tldr_language = get_language_code(os.environ.get('TLDR_LANGUAGE', ''))
     languages = os.environ.get('LANGUAGE', '').split(':')
     languages = list(map(
@@ -209,7 +229,12 @@ def get_language_list():
     return languages
 
 
-def get_page(command, remote=None, platforms=None, languages=None):
+def get_page(
+    command: str,
+    remote: Optional[str] = None,
+    platforms: Optional[List[str]] = None,
+    languages: Optional[List[str]] = None
+) -> Union[str, bool]:
     if platforms is None:
         platforms = get_platform_list()
     if languages is None:
@@ -271,11 +296,11 @@ ACCEPTED_COLOR_ATTRS = [
 
 LEADING_SPACES_NUM = 2
 
-COMMAND_SPLIT_REGEX = re.compile(r'(?P<param>{{.+?}})')
+COMMAND_SPLIT_REGEX = re.compile(r'(?P<param>{{.+?}*}})')
 PARAM_REGEX = re.compile(r'(?:{{)(?P<param>.+?)(?:}})')
 
 
-def get_commands(platforms=None):
+def get_commands(platforms: Optional[List[str]] = None) -> List[str]:
     if platforms is None:
         platforms = get_platform_list()
 
@@ -289,7 +314,7 @@ def get_commands(platforms=None):
     return commands
 
 
-def colors_of(key):
+def colors_of(key: str) -> Tuple[str, str, List[str]]:
     env_key = 'TLDR_COLOR_%s' % key.upper()
     values = os.environ.get(env_key, DEFAULT_COLORS[key]).strip().split()
     color = None
@@ -305,46 +330,49 @@ def colors_of(key):
     return (color, on_color, attrs)
 
 
-def output(page):
-    # Need a better fancy method?
-    if page is not None:
+def output(page: str, plain: bool = False) -> None:
+    if not plain:
         print()
-        for line in page:
-            line = line.rstrip().decode('utf-8')
-            if len(line) == 0:
-                continue
-            elif line[0] == '#':
-                line = ' ' * LEADING_SPACES_NUM + \
-                    colored(line.replace('# ', ''), *colors_of('name')) + '\n'
-                sys.stdout.buffer.write(line.encode('utf-8'))
-            elif line[0] == '>':
-                line = ' ' * (LEADING_SPACES_NUM - 1) + \
-                    colored(
-                        line.replace('>', '').replace('<', ''),
-                        *colors_of('description')
-                )
-                sys.stdout.buffer.write(line.encode('utf-8'))
-            elif line[0] == '-':
-                line = '\n' + ' ' * LEADING_SPACES_NUM + \
-                    colored(line, *colors_of('example'))
-                sys.stdout.buffer.write(line.encode('utf-8'))
-            elif line[0] == '`':
-                line = line[1:-1]  # need to actually parse ``
-                elements = [' ' * 2 * LEADING_SPACES_NUM]
-                for item in COMMAND_SPLIT_REGEX.split(line):
-                    item, replaced = PARAM_REGEX.subn(
-                        lambda x: colored(
-                            x.group('param'), *colors_of('parameter')),
-                        item)
-                    if not replaced:
-                        item = colored(item, *colors_of('command'))
-                    elements.append(item)
-                sys.stdout.buffer.write(''.join(elements).encode('utf-8'))
-            print()
+    for line in page:
+        line = line.rstrip().decode('utf-8')
+
+        if plain:
+            print(line)
+            continue
+        elif len(line) == 0:
+            continue
+        elif line[0] == '#':
+            line = ' ' * LEADING_SPACES_NUM + \
+                colored(line.replace('# ', ''), *colors_of('name')) + '\n'
+            sys.stdout.buffer.write(line.encode('utf-8'))
+        elif line[0] == '>':
+            line = ' ' * (LEADING_SPACES_NUM - 1) + \
+                colored(
+                    line.replace('>', '').replace('<', ''),
+                    *colors_of('description')
+            )
+            sys.stdout.buffer.write(line.encode('utf-8'))
+        elif line[0] == '-':
+            line = '\n' + ' ' * LEADING_SPACES_NUM + \
+                colored(line, *colors_of('example'))
+            sys.stdout.buffer.write(line.encode('utf-8'))
+        elif line[0] == '`':
+            line = line[1:-1]  # need to actually parse ``
+            elements = [' ' * 2 * LEADING_SPACES_NUM]
+            for item in COMMAND_SPLIT_REGEX.split(line):
+                item, replaced = PARAM_REGEX.subn(
+                    lambda x: colored(
+                        x.group('param'), *colors_of('parameter')),
+                    item)
+                if not replaced:
+                    item = colored(item, *colors_of('command'))
+                elements.append(item)
+            sys.stdout.buffer.write(''.join(elements).encode('utf-8'))
         print()
+    print()
 
 
-def update_cache(language=None):
+def update_cache(language: Optional[List[str]] = None) -> None:
     if language is None:
         tldr_language = os.environ.get("TLDR_LANGUAGE", get_default_language())
         language = tldr_language if tldr_language else 'en'
@@ -376,7 +404,7 @@ def update_cache(language=None):
         sys.exit("Error: Unable to update cache from " + DOWNLOAD_CACHE_LOCATION)
 
 
-def main():
+def create_parser() -> ArgumentParser:
     parser = ArgumentParser(
         prog="tldr",
         usage="tldr command [options]",
@@ -391,6 +419,11 @@ def main():
         )
     )
 
+    parser.add_argument("--search",
+                        metavar='"KEYWORDS"',
+                        type=str,
+                        help="Search for a specific command from a query")
+
     parser.add_argument('-u', '--update_cache',
                         action='store_true',
                         help="Update the local cache of pages and exit")
@@ -433,11 +466,30 @@ def main():
                         type=str,
                         help='Override the default language')
 
+    parser.add_argument('-m', '--markdown',
+                        default=False,
+                        action='store_true',
+                        help='Just print the plain page file.')
+
     parser.add_argument(
         'command', type=str, nargs='*', help="command to lookup", metavar='command'
-    ).completer = argcomplete.completers.ChoicesCompleter(get_commands())
+    ).complete = {"bash": "shtab_tldr_cmd_list", "zsh": "shtab_tldr_cmd_list"}
+
+    shtab.add_argument_to(parser, preamble={
+        'bash': r'''shtab_tldr_cmd_list(){{
+          compgen -W "$("{py}" -m tldr --list | sed 's/\W/ /g')" -- "$1"
+        }}'''.format(py=sys.executable),
+        'zsh': r'''shtab_tldr_cmd_list(){{
+          _describe 'command' "($("{py}" -m tldr --list | sed 's/\W/ /g'))"
+        }}'''.format(py=sys.executable)
+    })
+
+    return parser
+
+
+def main() -> None:
+    parser = create_parser()
 
-    argcomplete.autocomplete(parser)
     options = parser.parse_args()
 
     colorama.init(strip=options.color)
@@ -448,17 +500,56 @@ def main():
     elif len(sys.argv) == 1:
         parser.print_help(sys.stderr)
         sys.exit(1)
-
     if options.list:
         print(get_commands(options.platform))
     elif options.render:
         for command in options.command:
             if os.path.exists(command):
                 with open(command, encoding='utf-8') as open_file:
-                    output(open_file.read().encode('utf-8').splitlines())
+                    output(open_file.read().encode('utf-8').splitlines(),
+                           plain=options.markdown)
+    elif options.search:
+        command = '-'.join(options.command)
+        page = None
+        maxprob = 0
+        searchquery = options.search.split(' ')
+
+        platforms = get_platform_list()
+        for i in get_commands(platforms):
+            if i.startswith(command):
+                for p in platforms:
+                    result = load_page_from_cache(i, p, options.language)
+                    if result is not None and have_recent_cache(i, p, options.language):
+                        break
+                if result is None:
+                    raise SystemExit("Please update cache")
+                result = result.decode("utf-8")
+                result = result.split()
+                for word in searchquery:
+                    prob = 0
+                    for line in result:
+                        line = line.lower()
+                        if word in line:
+                            prob += 1
+                            if line.startswith("#"):
+                                prob += 7
+                            elif line.startswith(">"):
+                                prob += 5
+                            elif line.startswith("-"):
+                                prob += 2
+                    if prob > maxprob:
+                        maxprob = prob
+                        page = i
+
+        if page:
+            result = get_page(page, None, options.platform, options.language)
+            output(result, plain=options.markdown)
+        else:
+            print("No results found")
+
     else:
         try:
-            command = '-'.join(options.command)
+            command = '-'.join(options.command).lower()
             result = get_page(
                 command,
                 options.source,
@@ -472,12 +563,12 @@ def main():
                     " send a pull request to: https://github.com/tldr-pages/tldr"
                 ).format(cmd=command))
             else:
-                output(result)
+                output(result, plain=options.markdown)
         except URLError as e:
             sys.exit("Error fetching from tldr: {}".format(e))
 
 
-def cli():
+def cli() -> None:
     try:
         main()
     except KeyboardInterrupt:
